function FxDTSBrick::setPower(%this, %power, %source)
{
	%this.power = getMax(%power, 0);
	%this.source = %source;

	%this.setColorFX(%this.power <= 0 ? 0 : 3);
	%this.setColor(%this.power <= 0 ? 50 : 3);
}

function FxDTSBrick::receivePower(%this, %power, %source)
{
	if (%this.power < 0 && %power <= 0)
		return 0;

	// if (%this.power > 0 && %power > 0)
	// {
	// 	%power += %this.power;
	// }

	%vec = vectorSub(%this.getRayPos(), %source.getRayPos());
	%vec = vectorNormalize(%vec);

	%sum = getWord(%vec, 0) + getWord(%vec, 1) + getWord(%vec, 2);

	if (mAbs(%sum) != 1 || getWord(%vec, 2) != 0)
		return 0;

	switch$ (%vec)
	{
		case "1 0 0": %dir = 0;
		case "0 -1 0": %dir = 1;
		case "-1 0 0": %dir = 2;
		case "0 1 0": %dir = 3;
	}

	if (%dir $= "")
		return;

	%dir += %this.getDataBlock().orientationFix;
	%dir -= %this.angleID;

	if (%dir < 0)
		%dir += 4;

	if (%dir >= 4)
		%dir -= 4;

	%this.fireRelay(%this.client);

	if (%this.getDataBlock().isWire)
	{
		%this.setPower(%power, %source);
		PowerManager.push(%this);
		return 1;
	}

	if (%this.getDataBlock().isGate)
	{
		if (%dir == 0)
			return 0;

		%this.input[%dir] = %power;
		PowerManager.push(%this);
		return 1;
	}

	return 0;
}

function FxDTSBrick::processWire(%this)
{
	%db = %this.getDataBlock();

	if (%db.isGen)
		%power = %db.output;
	else if (%db.isWire)
		%power = %this.power;
	else
		return;

	%power -= 0.01;

	if (%power < 0)
		%power = 0;

	%links = %this.getPowerLinks();
	%links.remove(%this.source);

	%power /= %links.length;

	for (%i = 0; %i < %links.length; %i++)
		%links.value[%i].receivePower(%power, %this);

	%links.delete();
}

function FxDTSBrick::processGate(%this)
{
	%db = %this.getDataBlock();

	%power = 0;
	%count = 0;

	for (%i = 1; %i <= 3; %i++)
	{
		if (%this.input[%i] > 0)
			%count++;

		%power += %this.input[%i];
	}

	%logic = %db.processGate(%this, %count);
	%power *= mClampF(%logic, 0, 1);

	%this.output = %power;
	%power -= 0.01;

	%pos = %this.getRayPos();
	%dir = vectorScale(%this.getRayVec(), 0.5);

	%ray = containerRayCast(
		%pos, vectorAdd(%pos, %dir),
		$TypeMasks::FxBrickObjectType, %this
	);

	if (%ray)
		%ray.receivePower(%power, %this);
}

function PowerManager::onAdd(%this)
{
	%this.queue = Queue();

	%this.tickRate = 200;
	%this.tickQuota = 1;
}

function PowerManager::onRemove(%this)
{
	%this.queue.delete();
}

function PowerManager::push(%this, %object)
{
	if (!isObject(%object))
	{
		error("ERROR: Object '" @ %object @ "' does not exist");
		return 0;
	}

	if (!(%object.getType() & $TypeMasks::FxBrickObjectType))
	{
		error("ERROR: Object '" @ %object @ "' is not a brick");
		return 0;
	}

	if (!%object.getDataBlock().isWireLogic)
	{
		error("ERROR: Object '" @ %object @ "' is not a wire brick");
		return 0;
	}

	%this.queue.push(%object.getID());
	clientCmdBottomPrint("\c6PowerManager queue length: " @ %this.queue.length(), 1, 1);

	if (!isEventPending(%this.tickEvent))
		%this.tickEvent = %this.schedule(%this.tickRate, "tick");
}

function PowerManager::tick(%this)
{
	if (isEventPending(%this.tickEvent))
		cancel(%this.tickEvent);

	%count = 0;

	while (%this.queue.length() > 0 && %count < %this.tickQuota)
	{
		%object = %this.queue.pop();

		if (!isObject(%object))
			continue;

		%data = %object.getDataBlock();

		if (%data.isGate)
			%object.processGate(%this);
		else if (%data.isWire || %data.isGen)
			%object.processWire(%this);
		else
			continue;

		if (%this.showUpdate)
		{
			%projectile = new Projectile()
			{
				datablock = PongProjectile;
				scale = "0.25 0.25 0.25";

				initialPosition = %object.getWorldBoxCenter();
				initialVelocity = "0 0 0";
			};

			MissionCleanup.add(%projectile);
			%projectile.explode();
		}

		%count++;
	}

	clientCmdBottomPrint("\c6PowerManager queue length: " @ %this.queue.length(), 1, 1);

	if (%this.queue.length() > 0)
		%this.tickEvent = %this.schedule(%this.tickRate, "tick");
}

if (!isObject(PowerManager))
	new ScriptObject(PowerManager);